/** Definition of some bit vector management functions.
 *  @file vector.h
 *
 *  @author Renaud Lottiaux
 */

#ifndef __KRG_VECTOR__
#define __KRG_VECTOR__


#ifdef __KERNEL__
#include <linux/string.h>
#include <asm/bitops.h>
#else
#include <string.h>

/* Redefine the necessary bit functions since including the kernel header does
 * not work anymore from userspace */
/* All defs are taken from asm-i386/bitops.h and arch/i386/lib/bitops.c */

/* Assume SMP by default */
#define LOCK_PREFIX "\n\tlock; "

#define ADDR (*(volatile long *) addr)

static inline void set_bit(int nr, volatile unsigned long * addr)
{
	__asm__ __volatile__( LOCK_PREFIX
		"btsl %1,%0"
		:"+m" (ADDR)
		:"Ir" (nr));
}

static inline void clear_bit(int nr, volatile unsigned long * addr)
{
	__asm__ __volatile__( LOCK_PREFIX
		"btrl %1,%0"
		:"+m" (ADDR)
		:"Ir" (nr));
}

static inline int test_bit(int nr, const volatile unsigned long * addr)
{
	int oldbit;

	__asm__ __volatile__(
		"btl %2,%1\n\tsbbl %0,%0"
		:"=r" (oldbit)
		:"m" (ADDR),"Ir" (nr));
	return oldbit;
}

#undef ADDR

static inline unsigned long __ffs(unsigned long word)
{
	__asm__("bsfl %1,%0"
		:"=r" (word)
		:"rm" (word));
	return word;
}

static inline unsigned find_first_bit(const unsigned long *addr, unsigned size)
{
	unsigned x = 0;

	while (x < size) {
		unsigned long val = *addr++;
		if (val)
			return __ffs(val) + x;
		x += (sizeof(*addr)<<3);
	}
	return x;
}

/* Take from arch/i386/lib/bitops.c */
static inline int find_next_bit(const unsigned long *addr, int size, int offset)
{
	const unsigned long *p = addr + (offset >> 5);
	int set = 0, bit = offset & 31, res;

	if (bit) {
		/*
		 * Look for nonzero in the first 32 bits:
		 */
		__asm__("bsfl %1,%0\n\t"
			"jne 1f\n\t"
			"movl $32, %0\n"
			"1:"
			: "=r" (set)
			: "r" (*p >> bit));
		if (set < (32 - bit))
			return set + offset;
		set = 32 - bit;
		p++;
	}
	/*
	 * No set bit yet, search remaining full words for a bit
	 */
	res = find_first_bit (p, size - 32 * (p - addr));
	return (offset + set + res);
}

#endif /* __KERNEL__ */

#define KERRIGHED_VECTOR_SIZE_BITS 256
#define KERRIGHED_VECTOR_SIZE_BYTES (KERRIGHED_VECTOR_SIZE_BITS/8)

typedef struct vector {
  unsigned long vector[KERRIGHED_VECTOR_SIZE_BYTES/sizeof(unsigned long)];
} vector_t;

typedef vector_t kerrighed_nodes_t;

/** Operations on vectors */

/** Clear all bits in vector */
static inline void clear_vector(vector_t *vector)
{
  memset(vector->vector, 0, KERRIGHED_VECTOR_SIZE_BYTES);
}

/** Set the bit index in vector */
static inline void add_to_vector(vector_t *vector, int index)
{
  set_bit(index, vector->vector);
}

/** Reset the bit index in vector */
static inline void remove_from_vector(vector_t *vector, int index)
{
  clear_bit(index, vector->vector);
}

/** Return a non null number if the index is set in the vector */
static inline int is_in_vector(vector_t *vector, int index)
{
  return test_bit(index, vector->vector);
}

/** Return true if the index is the only one set in the vector */
static inline int is_unique_in_vector(vector_t *vector, int index)
{
  int i;
  
  i = find_first_bit(vector->vector, KERRIGHED_VECTOR_SIZE_BITS);
  if(i != index) return 0;
  
  i = find_next_bit(vector->vector, KERRIGHED_VECTOR_SIZE_BITS, index+1);
  if(i != KERRIGHED_VECTOR_SIZE_BITS) return 0;
  
  return 1;
}

/** Return true if the vector is empty */
static inline int vector_empty(vector_t *vector)
{
  return find_first_bit(vector->vector, KERRIGHED_VECTOR_SIZE_BITS) ==
    KERRIGHED_VECTOR_SIZE_BITS;
}

/** Copy a vector */
static inline void copy_vector(vector_t *vdest, vector_t *vsrc)
{
  memcpy(vdest->vector, vsrc->vector, KERRIGHED_VECTOR_SIZE_BYTES);
}

/** Find the first bit set in the vector */
static inline int find_first_in_vector(vector_t *vector)
{
  return find_first_bit(vector->vector, KERRIGHED_VECTOR_SIZE_BITS) ;
}

/** Find the next bit set in the vector */
static inline int find_next_in_vector(vector_t *vector, int index)
{
  return find_next_bit(vector->vector, KERRIGHED_VECTOR_SIZE_BITS, index) ;
}

static inline int is_equal_vector(vector_t *v1, vector_t *v2){
  int i;

  for(i=0;i<KERRIGHED_VECTOR_SIZE_BITS;i++)
    if( (is_in_vector(v1, i) && !is_in_vector(v2, i))
	|| (!is_in_vector(v1, i) && is_in_vector(v2, i))
	)
      return 0;

  return 1;
};

#ifdef __KERNEL__
#define print_vector(VEC) {int i; printk ("[ ");for(i=KERRIGHED_VECTOR_SIZE_BYTES/sizeof(unsigned long) - 1;i>=0;i--)printk("%08lx ", VEC.vector[i]); printk("]\n");}
#endif /* __KERNEL__ */

#endif // __KRG_VECTOR__
