/*
 *  Copyright (C) 2006-2007, Pascal Gallard, Kerlabs.
 */

#include <kerrighed/hashtable.h>

/*****************************************************************************/
/*                                                                           */
/*                           KERRIGHED HASH TABLES                           */
/*                                                                           */
/*****************************************************************************/


static struct hash_list HASH_LISTHEAD_NEW = { 0, NULL, NULL };


/** Add a new element in a hash table linked list.
 *  @author Viet Hoa Dinh
 * 
 *  The function must be called with the lock taken.
 *
 *  @param table  The table to add the element in.
 *  @param hash   The element key.
 *  @param data   The element to add in the table
 *
 *  @return  0 if everything ok.
 *           Negative value otherwise.
 */
static inline int hash_list_add(hashtable_t * table,
                                unsigned long hash,
				void * data)
{
  struct hash_list * ht;
  int index;

  index = hash % table->hashtable_size;

  ht = kmalloc(sizeof(struct hash_list), GFP_ATOMIC);
  if (ht == NULL)
    return -ENOMEM;

  ht->hash = hash;
  ht->data = data;
  ht->next = table->table[index].next;

  table->table[index].next = ht;

  return 0;
}



/** Remove an element from a hash table linked list.
 *  @author Viet Hoa Dinh
 *
 *  The function must be called with the lock taken.
 *
 *  @param table  The table to remove the element from.
 *  @param hash   The element key.
 *
 *  @return  0 if everything ok.
 *           Negative value otherwise.
 */
static inline void *hash_list_remove(hashtable_t * table,
				     unsigned long hash)
{
	struct hash_list * cur;
	void *data;
	int index;

	index = hash % table->hashtable_size;

	for(cur = &table->table[index] ; cur->next != NULL ; cur = cur->next) {
		if (cur->next->hash == hash) {
			struct hash_list * hash_data;

			hash_data = cur->next;
			data = hash_data->data;
			cur->next = cur->next->next;

			kfree(hash_data);
			return data;
		}
	}
	
	return NULL;
}



/** Free a hash table linked list.
 *  @author Viet Hoa Dinh
 *
 *  The function must be called with the lock taken.
 *
 *  @param list  List to free.
 */
static inline void hash_list_free(struct hash_list * list)
{
  struct hash_list * cur;
  struct hash_list * next;

  next = list;
  while (next != NULL) {
    cur = next;
    next = cur->next;
    kfree(cur);
  }
}



/** Find an element in a hash table linked list.
 *  @author Viet Hoa Dinh
 *
 *  The function must be called with the lock taken.
 *
 *  @param table  The table to find the element in.
 *  @param hash   The element key.
 *
 *  @return  A pointer to the found data.
 *           NULL if the data is not found.
 */
static inline void * hash_list_find(hashtable_t * table,
				    unsigned long hash)
{
  struct hash_list * cur;
  int index;

  index = hash % table->hashtable_size;

  for(cur = table->table[index].next ; cur != NULL ; cur = cur->next) {
    if (cur->hash == hash)
      return cur->data;
  }

  return NULL;
}


/** Create a new hash table */

hashtable_t * hashtable_new(unsigned long hashtable_size)
{
  hashtable_t * ht;
  int i;

  ht = kmalloc(sizeof(hashtable_t), GFP_KERNEL);
  if (ht == NULL)
    return NULL;

  ht->table = kmalloc(sizeof(struct hash_list) * hashtable_size, GFP_KERNEL);
  if (ht->table == NULL)
    return NULL;

  ht->hashtable_size = hashtable_size;
  spin_lock_init(&ht->lock);

  for(i = 0 ; i < hashtable_size ; i++)
    ht->table[i] = HASH_LISTHEAD_NEW;

  return ht;
}



/** Free a hash table */

void hashtable_free(hashtable_t * table)
{
  int i;
  unsigned long flags ;

  spin_lock_irqsave (&table->lock, flags);

  for(i = 0 ; i < table->hashtable_size ; i++)
    hash_list_free(table->table[i].next);

  spin_unlock_irqrestore (&table->lock, flags);

  kfree(table);
}



/** Add an element in a hash table */

int __hashtable_add(hashtable_t * table,
		    unsigned long hash,
		    void * data)
{
  int index;
  int r = 0 ;

#if defined(CONFIG_SMP)
  BUG_ON(!spin_is_locked(&table->lock));
#endif

  index = hash % table->hashtable_size;

  if (table->table[index].data == NULL)
    {
      table->table[index].hash = hash;
      table->table[index].data = data;
      table->table[index].next = NULL;
    }
  else
    r = hash_list_add(table, hash, data);

  return r ;
}


/** Remove an element from a hash table */

void *__hashtable_remove(hashtable_t * table,
			 unsigned long hash)
{
	int index;
	struct hash_list * next;
	void *data = NULL;
	
#if defined(CONFIG_SMP)
	BUG_ON(!spin_is_locked(&table->lock));
#endif
	
	index = hash % table->hashtable_size;
	
	if (table->table[index].hash == hash) {
		data = table->table[index].data;

		if ((next = table->table[index].next) != NULL) {
			table->table[index].hash = next->hash;
			table->table[index].data = next->data;
			table->table[index].next = next->next;
			kfree(next);
		}
		else {
			table->table[index].hash = 0;
			table->table[index].data = NULL;
		}
	}
	else
		data = hash_list_remove(table, hash);
	
	return data ;
}



/** Find an element in a hash table */

void * __hashtable_find(hashtable_t * table,
			unsigned long hash)
{
  int index;
  void * data ;


#if defined(CONFIG_SMP)
	BUG_ON(!spin_is_locked(&table->lock));
#endif

  index = hash % table->hashtable_size;

  if (table->table[index].hash == hash)
    data = table->table[index].data;
  else
    data = hash_list_find(table, hash);

  return data ;
}



/** Apply a function of each hash table key */

void __hashtable_foreach_key(hashtable_t * table,
			     void (* func)(int, void *),
			     void * data)
{
  unsigned long index;
  struct hash_list * cur;
  //unsigned long flags ;

  //spin_lock_irqsave (&table->lock, flags);

  for(index = 0 ; index < table->hashtable_size ; index ++) {
    if (table->table[index].data != NULL) {
      func(table->table[index].hash, data);
      for(cur = table->table[index].next ; cur != NULL ; cur = cur->next)
        func(cur->hash, data);
    }
  }

  //spin_unlock_irqrestore (&table->lock, flags);
}



/** Apply a function of each hash table element */

void __hashtable_foreach_data(hashtable_t * table,
			      void (* func)(void *, void *),
			      void * data)
{
  unsigned long index;
  struct hash_list * cur;
  //unsigned long flags ;

	//spin_lock_irqsave (&table->lock, flags);

  for(index = 0 ; index < table->hashtable_size ; index ++) {
    if (table->table[index].data != NULL) {
      func(table->table[index].data, data);
      for(cur = table->table[index].next ; cur != NULL ; cur = cur->next)
        func(cur->data, data);
    }
  }

	//spin_unlock_irqrestore (&table->lock, flags);

}

void * hashtable_find_data(hashtable_t * table,
			   int (* func)(void *, void *),
			   void * data)
{
  unsigned long index;
  struct hash_list * cur;
  unsigned long flags ;
  void * res = NULL ;

  spin_lock_irqsave (&table->lock, flags);

  for(index = 0 ; index < table->hashtable_size ; index ++) {
    if (table->table[index].data != NULL) {
      if (func(table->table[index].data, data))
	{
	  res = table->table[index].data ;
	  goto found ;
	}
      for(cur = table->table[index].next ; cur != NULL ; cur = cur->next)
        if (func(cur->data, data))
	  {
	    res = cur->data;
	    goto found;
	  }
    }
  }

found:
  spin_unlock_irqrestore (&table->lock, flags);
  return res ;
}
