/*
 *  Copyright (C) 2006-2007, Pascal Gallard, Kerlabs.
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/fs.h>

#include <kerrighed/types.h>
#include <kerrighed/krginit.h>
#include <kerrighed/krgsyms.h>
#include <kerrighed/unique_id.h>
#ifdef CONFIG_KRG_PROC
#include <kerrighed/pid.h>
#endif
#ifdef CONFIG_KRG_HOTPLUG
#include <kerrighed/hotplug.h>
#endif

/* Node id */
kerrighed_node_t kerrighed_node_id = 0;
EXPORT_SYMBOL(kerrighed_node_id);

/* Number of active nodes in the cluster */
kerrighed_node_t kerrighed_nb_nodes = 0;
EXPORT_SYMBOL(kerrighed_nb_nodes);

/* Min number of node before to start a cluster */
kerrighed_node_t kerrighed_nb_nodes_min = -1;
EXPORT_SYMBOL(kerrighed_nb_nodes_min);

/* Session id in order to mix several krg in the same physical network */
kerrighed_session_t kerrighed_session_id = 0;
EXPORT_SYMBOL(kerrighed_session_id);

/* ID of subcluster in the main one */
kerrighed_subsession_t kerrighed_subsession_id = -1;
EXPORT_SYMBOL(kerrighed_subsession_id);

/* Initialisation flags */
int kerrighed_init_flags = 0;
EXPORT_SYMBOL(kerrighed_init_flags);

/* lock around process transformation and hooks install */
DECLARE_RWSEM(kerrighed_init_sem);
EXPORT_SYMBOL(kerrighed_init_sem);

#ifdef CONFIG_KRG_PROC
/* Mask between local pid mode and global pid mode */
pid_t current_pid_mask = 0;
EXPORT_SYMBOL(current_pid_mask);
#endif


/*****************************************************************************/
/*                                                                           */
/*                          KERRIGHED INIT FUNCTION                          */
/*                                                                           */
/*****************************************************************************/

static char *read_from_file(char *_filename, int size)
{
	int error;
	struct file *f;
	char *b, *filename;

	b = kmalloc(size, GFP_ATOMIC);
	BUG_ON(b==NULL);

	filename = getname(_filename);
	if (!IS_ERR(filename)) {
		f = filp_open(filename, O_RDONLY, 0);
		if (IS_ERR(f)) {
			printk("error: %ld\n", PTR_ERR(f));
			goto err_file;
		}

		error = kernel_read(f, 0, b, size);
		//printk("read %d bytes\n", error);

		b[error] = 0;
		//printk(">>>%s<<<\n", b);

		if (f->f_op && f->f_op->flush) {
			error = f->f_op->flush(f, NULL);
			if (error)
				printk("global_pid_init: Error while closing file %d\n", error);
		}
	}
	return b;

 err_file:
	kfree(b);
	return NULL;
}

/* Remove then CR (if any) */
static void strip_hostname(char *h)
{
	char *i;

	for (i = h; *i; i++) {
		if (*i == 10) {
			*i=0;
			break;
		}
	}
}

static char *get_next_line(char *k)
{
	char *i;

	BUG_ON(*k==0);

	for (i = k; *i; i++) {
		if (*i == 10)
			return i+1;
	}

	return NULL;
}

static void read_kerrighed_nodes(char *_h, char *k)
{
	char *ik, *h;
	int lh;

	if ((_h==NULL) || (k==NULL))
		return;

	lh = strlen(_h);
	h = kmalloc(lh+1, GFP_ATOMIC);
	strncpy(h, _h, lh);
	h[lh] = ':';
	h[lh+1] = 0;
	lh = strlen(h);

	kerrighed_nb_nodes = 0;
	for (ik=k; ik && *ik;) {
		if (!ISSET_KRG_INIT_FLAGS(KRG_INITFLAGS_SESSIONID)) {
			if (strncmp("session=", ik, 8) == 0){
				ik += 8;
				kerrighed_session_id = simple_strtoul(ik, NULL, 10);
				SET_KRG_INIT_FLAGS(KRG_INITFLAGS_SESSIONID);

				ik = get_next_line(ik);
				continue;
			}
		}

		if (strncmp("nbmin=", ik, 6) == 0) {
			ik += 6;
			kerrighed_nb_nodes_min = simple_strtoul(ik, NULL, 10);
			ik = get_next_line(ik);
			continue;
		}

		if (!ISSET_KRG_INIT_FLAGS(KRG_INITFLAGS_NODEID)) {
			if (strncmp(h, ik, lh) == 0) {
				char *end;
				ik += lh;

				kerrighed_node_id = simple_strtoul(ik, &end, 10);
				SET_KRG_INIT_FLAGS(KRG_INITFLAGS_NODEID);
			}
		}

		if (!ISSET_KRG_INIT_FLAGS(KRG_INITFLAGS_NBNODES))
			kerrighed_nb_nodes++;

		ik = get_next_line(ik);
	}

	if (!ISSET_KRG_INIT_FLAGS(KRG_INITFLAGS_NBNODES))
		SET_KRG_INIT_FLAGS(KRG_INITFLAGS_NBNODES);
}

void global_pid_init(int nodeid, int nbnodes,int session_id, int nbmin)
{
	char *hostname, *kerrighed_nodes;
	int need_read = 0;

	if (!ISSET_KRG_INIT_FLAGS(KRG_INITFLAGS_NODEID)) {
		if (nodeid >= 0) {
			kerrighed_node_id = nodeid;
			SET_KRG_INIT_FLAGS(KRG_INITFLAGS_NODEID);
		} else
			need_read = 1;
	}

	if (nbnodes >= 0) {
		kerrighed_nb_nodes = nbnodes;
		SET_KRG_INIT_FLAGS(KRG_INITFLAGS_NBNODES);
	} else
		need_read = 1;

	if (session_id >= 0) {
		kerrighed_session_id = session_id;
		SET_KRG_INIT_FLAGS(KRG_INITFLAGS_SESSIONID);
	} else
		need_read = 1;

	if (nbmin >= 0)
		kerrighed_nb_nodes_min = nbmin;

	if (need_read) {
		/* first we read the name of the node */
		hostname = read_from_file("/etc/hostname", 256);
		if (!hostname) {
			printk("Can't read /etc/hostname\n");
			goto out;
		}
		strip_hostname(hostname);

		kerrighed_nodes = read_from_file("/etc/kerrighed_nodes", 4096);
		if (!kerrighed_nodes) {
			kfree(hostname);
			printk("Can't read /etc/kerrighed_nodes\n");
			goto out;
		}
		read_kerrighed_nodes(hostname, kerrighed_nodes);

		kfree(kerrighed_nodes);
		kfree(hostname);
	}

 out:
	if (ISSET_KRG_INIT_FLAGS(KRG_INITFLAGS_NODEID)) {
#ifdef CONFIG_KRG_HOTPLUG
		universe[kerrighed_node_id].state = 1;
#endif
#ifdef CONFIG_KRG_PROC
		current_pid_mask = GLOBAL_PID_MASK;
#endif
	}

	printk("Kerrighed session ID : %d\n", kerrighed_session_id);
	printk("Kerrighed node ID    : %d\n", kerrighed_node_id);
	printk("Kerrighed nb nodes   : %d\n", kerrighed_nb_nodes);
	printk("Kerrighed min nodes  : %d\n", kerrighed_nb_nodes_min);

	return;
}

EXPORT_SYMBOL(global_pid_init);

void init_node_discovering(void);

void kerrighed_init(void)
{
	init_krgsyms();
	init_unique_ids();
	init_node_discovering();
}
